
#include "spdcc.h"

/*--------------------------------------------------------------------------*/
/* Force external symbol creation for inline functions */
/*--------------------------------------------------------------------------*/

extern inline int spdc_ravel(const spdc_mdarray * arr,
			     const ptrdiff_t * multi_idx, ptrdiff_t * flat_idx);
extern inline int spdc_unravel(const spdc_mdarray * arr, ptrdiff_t flat_idx,
			       ptrdiff_t * multi_idx);

extern inline double spdc_get_double(const spdc_mdarray * arr, ptrdiff_t * n);
extern inline double complex spdc_get_complex(const spdc_mdarray * arr,
					      ptrdiff_t * n);

/*----------------------------------------------------------------------------
 * Domain-specific logic such as Gauss-Hermite modes and refractive indices of
 * BBO
 *-------------------------------------------------------------------------- */

int spdc_gauss_hermite(int m, int n, spdc_mdarray * x, spdc_mdarray * y,
		       double z, double w_0, double lambda, spdc_mdarray * res)
{
	double z_r, w, R, xi, k, x_tmp, y_tmp;
	size_t N = 0;
	ptrdiff_t j[2];
	double complex exp_term, h_term, *data = NULL;

	/* Validate input array shapes etc.. */
	if (res->dtype != SPDC_DTYPE_COMPLEX) {
		return (SPDC_WRONG_DTYPE);
	}

	if ((x->dtype != SPDC_DTYPE_DOUBLE) | (y->dtype != SPDC_DTYPE_DOUBLE)) {
		return (SPDC_WRONG_DTYPE);
	}

	if ((res->ndim != 2) | (x->ndim != 1) | (y->ndim != 1)) {
		return (SPDC_NDIM_INCOMPATIBLE);
	}

	if ((res->shape[0] != x->shape[0]) | (res->shape[1] != (y->shape[0]))) {
		return (SPDC_SHAPE_INCOMPATIBLE);
	}

	/* Calculate rayleigh range and curvature etc. */
	if (z == 0.0) {
		z = 1e-12;
	}

	z_r = M_PI * pow(w_0, 2) / lambda;
	w = w_0 * sqrt(1.0 + pow(z / z_r, 2));
	R = z * (1.0 + pow(z_r / z, 2));
	xi = atan(z / z_r);
	k = 2.0 * M_PI / lambda;

	/* Fill output array */
	N = spdc_get_size(res);
	assert(N != 0);
	data = (double complex *)res->data;
	for (ptrdiff_t i = 0; i < N; i++) {
		spdc_unravel(res, i, j);
		x_tmp = spdc_get_double(x, &(j[0]));
		y_tmp = spdc_get_double(y, &(j[1]));

		exp_term =
		    cexp(-(pow(x_tmp, 2) + pow(y_tmp, 2)) *
			 (1.0 / pow(w, 2) + I * k / (2.0 * R)) - I * k * z -
			 I * (m + n + 1) * xi);

		h_term = spdc_hermite_H(m, sqrt(2.0) * x_tmp / w);
		h_term *= spdc_hermite_H(n, sqrt(2.0) * y_tmp / w);
		data[i] = w_0 / w * h_term * exp_term;
	}

	return (SPDC_RETURN_OK);
}

int spdc_pseudo_tem10(spdc_mdarray * x, spdc_mdarray * y, double w_0,
		      double lambda, spdc_mdarray * res)
{
	/* Calculate the near-field field distribution of an electric field that
	 * was created by applying a phase shift of pi to one half of the
	 * far-field electric field distribution
	 *
	 *  :width:     The width of the zone that is blackened (in the far field)
	 *              between the two lobes
	 */
	size_t N = spdc_get_size(res);
	size_t norm_2d = res->shape[0] * res->shape[1];
	fftw_plan pf, pb;

	pf = fftw_plan_dft_2d((int)res->shape[0], (int)res->shape[1], res->data,
			      res->data, FFTW_FORWARD, FFTW_ESTIMATE);
	pb = fftw_plan_dft_2d((int)res->shape[0], (int)res->shape[1], res->data,
			      res->data, FFTW_BACKWARD, FFTW_ESTIMATE);

	spdc_gauss_hermite(0, 0, x, y, 0.0, w_0, lambda, res);

	/* Calculate far-field pattern and apply phase mask */
	fftw_execute(pf);
	for (ptrdiff_t flat_idx = 0; flat_idx < N; flat_idx++) {
		ptrdiff_t i, j;
		SPDC_UNRAVEL2(*res, flat_idx, i, j);
		if (i < (res->shape[0] / 2.0)) {
			SPDC_GET_COMPLEX(*res, flat_idx) *= -1.0;
		}
		if (i == 0) {
			SPDC_GET_COMPLEX(*res, flat_idx) = 0.0;
		}
	}

	/* Transform back to near-field */
	fftw_execute(pb);
	for (ptrdiff_t i = 0; i < N; i++) {
		SPDC_GET_COMPLEX(*res, i) /= norm_2d;
	}
	return (SPDC_RETURN_OK);

}

int spdc_walkoff_coefficients(double n_o, double n_e, double theta,
			      field_co * co)
{
	/*Calculate the four coefficients for the Fresnel wave propagation
	   integral alpha, beta, gamma and eta as defined in [arXiv:1010.1236] Eqs.
	   49-52.

	   :param theta:   Angle between the surface normal (or z-axis) of the
	   nonlinear crystal and its optical axis (OA).
	   :param n_e:     The extraordinary refractive index of the crystal
	   :param n_o:     The ordinary refractive index of the crystal

	   :returns:       alpha, beta, gamma, eta coefficients
	 */

	/* Validate input values */
	assert(n_o >= 1.0);
	assert(n_e >= 1.0);

	double den_1, den_2, alpha, beta, gamma, eta;
	den_1 =
	    pow(n_o, 2) * pow(sin(theta), 2) + pow(n_e, 2) * pow(cos(theta), 2);
	den_2 = sqrt(den_1);
	alpha = (pow(n_o, 2) - pow(n_e, 2)) * sin(theta) * cos(theta) / den_1;
	beta = n_o * n_e / den_1;
	gamma = n_o / den_2;
	eta = n_o * n_e / den_2;
	co->alpha = alpha;
	co->beta = beta;
	co->gamma = gamma;
	co->eta = eta;
	return (SPDC_RETURN_OK);
}

int spdc_sellmeier_bbo(double l, double *n_o, double *n_e)
{
	/* Input contract */
	assert(n_o);
	assert(n_e);

	l *= 1e6;
	double l_2 = pow(l, 2);
	*n_o = sqrt(2.7359 + 0.01878 / (l_2 - 0.01822) - 0.01354 * l_2);
	*n_e = sqrt(2.3753 + 0.01224 / (l_2 - 0.01667) - 0.01516 * l_2);
	return (SPDC_RETURN_OK);
}

/* Calculate the physicists hermite polynomials H_n(x) for n <= 4 */
double complex spdc_hermite_H(int n, double complex x)
{
	assert(n <= 4);

	double complex res = 0;
	if (n == 0) {
		res = 1.0;
	} else if (n == 1) {
		res = 2.0 * x;
	} else if (n == 2) {
		res = 4.0 * x * x - 2.0;
	} else if (n == 3) {
		res = (8.0 * x * x - 12.0) * x;
	} else if (n == 4) {
		res = 12.0 + x * x * (16.0 * x * x - 48.0);
	}

	return (res);
}

/*-----------------------------------------------------------------------------
 * Library utility functions such as array access and initialization
 *---------------------------------------------------------------------------*/

int spdc_dump_prm(struct spdc_prm prm)
{
	printf("# This version of SPDCc is: %s\n", VERSION);
	printf("# The current parameters are\n");
	/*[[[cog
	   import cog_codegen

	   options = cog_codegen.get_options()
	   cog_codegen.generate_printing_code(options)
	   ]]] */
	printf("# pump beam vacuum wavelength (lambda-pump): %e\n",
	       prm.lambda_pump);
	printf("# signal beam vacuum wavelength (lambda-signal): %e\n",
	       prm.lambda_signal);
	printf("# idler beam vacuum wavelength (lambda-idler): %e\n",
	       prm.lambda_idler);
	printf
	    ("# m order of the hermite-gauss TEM_mn pump mode (hermite-gauss-m): %d\n",
	     prm.hermite_gauss_m);
	printf
	    ("# n order of the hermite-gauss TEM_mn pump mode (hermite-gauss-n): %d\n",
	     prm.hermite_gauss_n);
	printf("# w0 of the pump beam (pump-w0): %e\n", prm.pump_w0);
	printf("# z of the pump beam (pump-z): %e\n", prm.pump_z);
	printf
	    ("# Use a pseuo-TEM mode generated from an TEM via a mode converter (pseudo-tem): %d\n",
	     prm.pseudo_tem);
	printf("# Transverse crystal width (width): %e\n", prm.width);
	printf("# Crystal length in z-direction (length): %e\n", prm.length);
	printf
	    ("# Angle of the crystal's OA wrt to the z-axis in degree (theta): %e\n",
	     prm.theta);
	printf("# Enable exponential phase matching term (phase): %d\n",
	       prm.phase);
	printf("# Enable exp phase compensation (phase-compensation): %d\n",
	       prm.phase_compensation);
	printf
	    ("# Set phase II (default) or I phase matching (phase-match-type): %d\n",
	     prm.phase_match_type);
	printf("# Number of discrete samples in each dimension (N): %d\n",
	       prm.N);
	printf
	    ("# Rank of the approximate Schmidt decomposition (schmidt-rank): %d\n",
	     prm.schmidt_rank);
	printf("# Only output the eigenvalues S (eigenvalues-only): %d\n",
	       prm.eigenvalues_only);
	printf
	    ("# Focal length of the lenses of the 4F correlator (focal-length): %e\n",
	     prm.focal_length);
	printf
	    ("# Longitudinal near field signal offset (nf-signal-offset): %e\n",
	     prm.nf_signal_offset);
	printf("# Longitudinal near field idler offset (nf-idler-offset): %e\n",
	       prm.nf_idler_offset);
	printf
	    ("# Longitudinal far field signal offset (ff-signal-offset): %e\n",
	     prm.ff_signal_offset);
	printf("# Longitudinal far field idler offset (ff-idler-offset): %e\n",
	       prm.ff_idler_offset);
	printf
	    ("# A knife edge aperture at pump beam plane pump-pump in x-direction (knife-edge-pos-pump-px): %e\n",
	     prm.knife_edge_pos_pump_px);
	printf
	    ("# A knife edge aperture at pump beam plane pump-pump in y-direction (knife-edge-pos-pump-py): %e\n",
	     prm.knife_edge_pos_pump_py);
	printf
	    ("# A knife edge aperture at pump beam plane pump-pump in x-direction (knife-edge-invert-pump-px): %d\n",
	     prm.knife_edge_invert_pump_px);
	printf
	    ("# A knife edge aperture at pump beam plane pump-pump in y-direction (knife-edge-invert-pump-py): %d\n",
	     prm.knife_edge_invert_pump_py);
	printf
	    ("# circular aperture mask at pump beam plane pump-pump (circ-ap-r-pump-p): %e\n",
	     prm.circ_ap_r_pump_p);
	printf
	    ("# circular aperture mask at pump beam plane pump-pump (circ-ap-cntr-x-pump-p): %e\n",
	     prm.circ_ap_cntr_x_pump_p);
	printf
	    ("# circular aperture mask at pump beam plane pump-pump (circ-ap-cntr-y-pump-p): %e\n",
	     prm.circ_ap_cntr_y_pump_p);
	printf
	    ("# a single-slit aperture at pump beam plane pump-pump in x-direction (single-slit-cntr-pump-px): %e\n",
	     prm.single_slit_cntr_pump_px);
	printf
	    ("# a single-slit aperture at pump beam plane pump-pump in y-direction (single-slit-cntr-pump-py): %e\n",
	     prm.single_slit_cntr_pump_py);
	printf
	    ("# a single-slit aperture at pump beam plane pump-pump in x-direction (single-slit-width-pump-px): %e\n",
	     prm.single_slit_width_pump_px);
	printf
	    ("# a single-slit aperture at pump beam plane pump-pump in y-direction (single-slit-width-pump-py): %e\n",
	     prm.single_slit_width_pump_py);
	printf
	    ("# a double-slit aperture at pump beam plane pump-pump in x-direction (double-slit-sep-pump-px): %e\n",
	     prm.double_slit_sep_pump_px);
	printf
	    ("# a double-slit aperture at pump beam plane pump-pump in y-direction (double-slit-sep-pump-py): %e\n",
	     prm.double_slit_sep_pump_py);
	printf
	    ("# a double-slit aperture at pump beam plane pump-pump in x-direction (double-slit-width-pump-px): %e\n",
	     prm.double_slit_width_pump_px);
	printf
	    ("# a double-slit aperture at pump beam plane pump-pump in y-direction (double-slit-width-pump-py): %e\n",
	     prm.double_slit_width_pump_py);
	printf
	    ("# A knife edge aperture at far-field SPDC plane f1-signal in x-direction (knife-edge-pos-f1-sx): %e\n",
	     prm.knife_edge_pos_f1_sx);
	printf
	    ("# A knife edge aperture at far-field SPDC plane f1-signal in y-direction (knife-edge-pos-f1-sy): %e\n",
	     prm.knife_edge_pos_f1_sy);
	printf
	    ("# A knife edge aperture at far-field SPDC plane f1-signal in x-direction (knife-edge-invert-f1-sx): %d\n",
	     prm.knife_edge_invert_f1_sx);
	printf
	    ("# A knife edge aperture at far-field SPDC plane f1-signal in y-direction (knife-edge-invert-f1-sy): %d\n",
	     prm.knife_edge_invert_f1_sy);
	printf
	    ("# circular aperture mask at far-field SPDC plane f1-signal (circ-ap-r-f1-s): %e\n",
	     prm.circ_ap_r_f1_s);
	printf
	    ("# circular aperture mask at far-field SPDC plane f1-signal (circ-ap-cntr-x-f1-s): %e\n",
	     prm.circ_ap_cntr_x_f1_s);
	printf
	    ("# circular aperture mask at far-field SPDC plane f1-signal (circ-ap-cntr-y-f1-s): %e\n",
	     prm.circ_ap_cntr_y_f1_s);
	printf
	    ("# a single-slit aperture at far-field SPDC plane f1-signal in x-direction (single-slit-cntr-f1-sx): %e\n",
	     prm.single_slit_cntr_f1_sx);
	printf
	    ("# a single-slit aperture at far-field SPDC plane f1-signal in y-direction (single-slit-cntr-f1-sy): %e\n",
	     prm.single_slit_cntr_f1_sy);
	printf
	    ("# a single-slit aperture at far-field SPDC plane f1-signal in x-direction (single-slit-width-f1-sx): %e\n",
	     prm.single_slit_width_f1_sx);
	printf
	    ("# a single-slit aperture at far-field SPDC plane f1-signal in y-direction (single-slit-width-f1-sy): %e\n",
	     prm.single_slit_width_f1_sy);
	printf
	    ("# a double-slit aperture at far-field SPDC plane f1-signal in x-direction (double-slit-sep-f1-sx): %e\n",
	     prm.double_slit_sep_f1_sx);
	printf
	    ("# a double-slit aperture at far-field SPDC plane f1-signal in y-direction (double-slit-sep-f1-sy): %e\n",
	     prm.double_slit_sep_f1_sy);
	printf
	    ("# a double-slit aperture at far-field SPDC plane f1-signal in x-direction (double-slit-width-f1-sx): %e\n",
	     prm.double_slit_width_f1_sx);
	printf
	    ("# a double-slit aperture at far-field SPDC plane f1-signal in y-direction (double-slit-width-f1-sy): %e\n",
	     prm.double_slit_width_f1_sy);
	printf
	    ("# A knife edge aperture at far-field SPDC plane f1-idler in x-direction (knife-edge-pos-f1-ix): %e\n",
	     prm.knife_edge_pos_f1_ix);
	printf
	    ("# A knife edge aperture at far-field SPDC plane f1-idler in y-direction (knife-edge-pos-f1-iy): %e\n",
	     prm.knife_edge_pos_f1_iy);
	printf
	    ("# A knife edge aperture at far-field SPDC plane f1-idler in x-direction (knife-edge-invert-f1-ix): %d\n",
	     prm.knife_edge_invert_f1_ix);
	printf
	    ("# A knife edge aperture at far-field SPDC plane f1-idler in y-direction (knife-edge-invert-f1-iy): %d\n",
	     prm.knife_edge_invert_f1_iy);
	printf
	    ("# circular aperture mask at far-field SPDC plane f1-idler (circ-ap-r-f1-i): %e\n",
	     prm.circ_ap_r_f1_i);
	printf
	    ("# circular aperture mask at far-field SPDC plane f1-idler (circ-ap-cntr-x-f1-i): %e\n",
	     prm.circ_ap_cntr_x_f1_i);
	printf
	    ("# circular aperture mask at far-field SPDC plane f1-idler (circ-ap-cntr-y-f1-i): %e\n",
	     prm.circ_ap_cntr_y_f1_i);
	printf
	    ("# a single-slit aperture at far-field SPDC plane f1-idler in x-direction (single-slit-cntr-f1-ix): %e\n",
	     prm.single_slit_cntr_f1_ix);
	printf
	    ("# a single-slit aperture at far-field SPDC plane f1-idler in y-direction (single-slit-cntr-f1-iy): %e\n",
	     prm.single_slit_cntr_f1_iy);
	printf
	    ("# a single-slit aperture at far-field SPDC plane f1-idler in x-direction (single-slit-width-f1-ix): %e\n",
	     prm.single_slit_width_f1_ix);
	printf
	    ("# a single-slit aperture at far-field SPDC plane f1-idler in y-direction (single-slit-width-f1-iy): %e\n",
	     prm.single_slit_width_f1_iy);
	printf
	    ("# a double-slit aperture at far-field SPDC plane f1-idler in x-direction (double-slit-sep-f1-ix): %e\n",
	     prm.double_slit_sep_f1_ix);
	printf
	    ("# a double-slit aperture at far-field SPDC plane f1-idler in y-direction (double-slit-sep-f1-iy): %e\n",
	     prm.double_slit_sep_f1_iy);
	printf
	    ("# a double-slit aperture at far-field SPDC plane f1-idler in x-direction (double-slit-width-f1-ix): %e\n",
	     prm.double_slit_width_f1_ix);
	printf
	    ("# a double-slit aperture at far-field SPDC plane f1-idler in y-direction (double-slit-width-f1-iy): %e\n",
	     prm.double_slit_width_f1_iy);
	printf
	    ("# A knife edge aperture at near-field SPDC plane f2-signal in x-direction (knife-edge-pos-f2-sx): %e\n",
	     prm.knife_edge_pos_f2_sx);
	printf
	    ("# A knife edge aperture at near-field SPDC plane f2-signal in y-direction (knife-edge-pos-f2-sy): %e\n",
	     prm.knife_edge_pos_f2_sy);
	printf
	    ("# A knife edge aperture at near-field SPDC plane f2-signal in x-direction (knife-edge-invert-f2-sx): %d\n",
	     prm.knife_edge_invert_f2_sx);
	printf
	    ("# A knife edge aperture at near-field SPDC plane f2-signal in y-direction (knife-edge-invert-f2-sy): %d\n",
	     prm.knife_edge_invert_f2_sy);
	printf
	    ("# circular aperture mask at near-field SPDC plane f2-signal (circ-ap-r-f2-s): %e\n",
	     prm.circ_ap_r_f2_s);
	printf
	    ("# circular aperture mask at near-field SPDC plane f2-signal (circ-ap-cntr-x-f2-s): %e\n",
	     prm.circ_ap_cntr_x_f2_s);
	printf
	    ("# circular aperture mask at near-field SPDC plane f2-signal (circ-ap-cntr-y-f2-s): %e\n",
	     prm.circ_ap_cntr_y_f2_s);
	printf
	    ("# a single-slit aperture at near-field SPDC plane f2-signal in x-direction (single-slit-cntr-f2-sx): %e\n",
	     prm.single_slit_cntr_f2_sx);
	printf
	    ("# a single-slit aperture at near-field SPDC plane f2-signal in y-direction (single-slit-cntr-f2-sy): %e\n",
	     prm.single_slit_cntr_f2_sy);
	printf
	    ("# a single-slit aperture at near-field SPDC plane f2-signal in x-direction (single-slit-width-f2-sx): %e\n",
	     prm.single_slit_width_f2_sx);
	printf
	    ("# a single-slit aperture at near-field SPDC plane f2-signal in y-direction (single-slit-width-f2-sy): %e\n",
	     prm.single_slit_width_f2_sy);
	printf
	    ("# a double-slit aperture at near-field SPDC plane f2-signal in x-direction (double-slit-sep-f2-sx): %e\n",
	     prm.double_slit_sep_f2_sx);
	printf
	    ("# a double-slit aperture at near-field SPDC plane f2-signal in y-direction (double-slit-sep-f2-sy): %e\n",
	     prm.double_slit_sep_f2_sy);
	printf
	    ("# a double-slit aperture at near-field SPDC plane f2-signal in x-direction (double-slit-width-f2-sx): %e\n",
	     prm.double_slit_width_f2_sx);
	printf
	    ("# a double-slit aperture at near-field SPDC plane f2-signal in y-direction (double-slit-width-f2-sy): %e\n",
	     prm.double_slit_width_f2_sy);
	printf
	    ("# A knife edge aperture at near-field SPDC plane f2-idler in x-direction (knife-edge-pos-f2-ix): %e\n",
	     prm.knife_edge_pos_f2_ix);
	printf
	    ("# A knife edge aperture at near-field SPDC plane f2-idler in y-direction (knife-edge-pos-f2-iy): %e\n",
	     prm.knife_edge_pos_f2_iy);
	printf
	    ("# A knife edge aperture at near-field SPDC plane f2-idler in x-direction (knife-edge-invert-f2-ix): %d\n",
	     prm.knife_edge_invert_f2_ix);
	printf
	    ("# A knife edge aperture at near-field SPDC plane f2-idler in y-direction (knife-edge-invert-f2-iy): %d\n",
	     prm.knife_edge_invert_f2_iy);
	printf
	    ("# circular aperture mask at near-field SPDC plane f2-idler (circ-ap-r-f2-i): %e\n",
	     prm.circ_ap_r_f2_i);
	printf
	    ("# circular aperture mask at near-field SPDC plane f2-idler (circ-ap-cntr-x-f2-i): %e\n",
	     prm.circ_ap_cntr_x_f2_i);
	printf
	    ("# circular aperture mask at near-field SPDC plane f2-idler (circ-ap-cntr-y-f2-i): %e\n",
	     prm.circ_ap_cntr_y_f2_i);
	printf
	    ("# a single-slit aperture at near-field SPDC plane f2-idler in x-direction (single-slit-cntr-f2-ix): %e\n",
	     prm.single_slit_cntr_f2_ix);
	printf
	    ("# a single-slit aperture at near-field SPDC plane f2-idler in y-direction (single-slit-cntr-f2-iy): %e\n",
	     prm.single_slit_cntr_f2_iy);
	printf
	    ("# a single-slit aperture at near-field SPDC plane f2-idler in x-direction (single-slit-width-f2-ix): %e\n",
	     prm.single_slit_width_f2_ix);
	printf
	    ("# a single-slit aperture at near-field SPDC plane f2-idler in y-direction (single-slit-width-f2-iy): %e\n",
	     prm.single_slit_width_f2_iy);
	printf
	    ("# a double-slit aperture at near-field SPDC plane f2-idler in x-direction (double-slit-sep-f2-ix): %e\n",
	     prm.double_slit_sep_f2_ix);
	printf
	    ("# a double-slit aperture at near-field SPDC plane f2-idler in y-direction (double-slit-sep-f2-iy): %e\n",
	     prm.double_slit_sep_f2_iy);
	printf
	    ("# a double-slit aperture at near-field SPDC plane f2-idler in x-direction (double-slit-width-f2-ix): %e\n",
	     prm.double_slit_width_f2_ix);
	printf
	    ("# a double-slit aperture at near-field SPDC plane f2-idler in y-direction (double-slit-width-f2-iy): %e\n",
	     prm.double_slit_width_f2_iy);
	printf("# Mask type at plane pump-pump (pump-pump-mask-type): %d\n",
	       prm.pump_pump_mask_type);
	printf("# Mask type at plane f1-signal (f1-signal-mask-type): %d\n",
	       prm.f1_signal_mask_type);
	printf("# Mask type at plane f1-idler (f1-idler-mask-type): %d\n",
	       prm.f1_idler_mask_type);
	printf("# Mask type at plane f2-signal (f2-signal-mask-type): %d\n",
	       prm.f2_signal_mask_type);
	printf("# Mask type at plane f2-idler (f2-idler-mask-type): %d\n",
	       prm.f2_idler_mask_type);
	/*[[[end]]] */

	return (SPDC_RETURN_OK);
}

int spdc_dump_array(spdc_mdarray * arr)
{
	size_t N = spdc_get_size(arr);

	if (arr->dtype == SPDC_DTYPE_COMPLEX) {
		for (ptrdiff_t i = 0; i < N; i++) {
			double complex current_item =
			    ((double complex *)arr->data)[i];
			printf("%e\t%e\n", creal(current_item),
			       cimag(current_item));
		}
	} else if (arr->dtype == SPDC_DTYPE_DOUBLE) {
		for (ptrdiff_t i = 0; i < N; i++) {
			printf("%e\n", ((double *)arr->data)[i]);
		}
	} else {
		return (SPDC_WRONG_DTYPE);
	}

	return (SPDC_RETURN_OK);
}

int spdc_fftshift(spdc_mdarray * arr)
{
	if (arr->ndim != 1) {
		return (SPDC_NDIM_INCOMPATIBLE);
	}

	if ((arr->shape[0] % 2) != 0) {
		return (SPDC_NOT_POWER_OF_TWO);
	}

	size_t n_bytes = 0, N = 0;
	N = spdc_get_size(arr);
	if (arr->dtype == SPDC_DTYPE_DOUBLE) {
		n_bytes = N * sizeof(double);
	} else if (arr->dtype == SPDC_DTYPE_COMPLEX) {
		n_bytes = N * sizeof(double complex);
	} else {
		return (SPDC_WRONG_DTYPE);
	}

	/* Copy memory */
	void *tmp_storage = NULL;
	tmp_storage = malloc(n_bytes / 2);
	if (!tmp_storage) {
		return (SPDC_MEMORY_ALLOC_ERROR);
	}

	memcpy(tmp_storage, arr->data, n_bytes / 2);
	memcpy(arr->data, (arr->data + n_bytes / 2), n_bytes / 2);
	memcpy((arr->data + n_bytes / 2), tmp_storage, n_bytes / 2);
	free(tmp_storage);
	return (SPDC_RETURN_OK);
}

int spdc_fftfreq(spdc_mdarray * arr, double sampling_period)
{
	if (arr->ndim != 1) {
		return (SPDC_NDIM_INCOMPATIBLE);
	}

	if (arr->dtype != SPDC_DTYPE_DOUBLE) {
		return (SPDC_WRONG_DTYPE);
	}

	ptrdiff_t i;
	size_t n_items = spdc_get_size(arr);
	double *data = (double *)arr->data;

	for (i = 0; i < n_items; i++) {
		if (i < n_items / 2.0) {
			data[i] = ((double)i) / sampling_period;
		} else {
			data[i] =
			    -((double)n_items - (double)i) / sampling_period;
		}
	}
	return (SPDC_RETURN_OK);
}

/* Fill a one-dimensional mdarray with N evenly spaced samples from min to
 * max. The number of sample points in the range [min, max] is determined by
 * the shape of the array. The array is modified in place. The data type of
 * the array must be double. If a wrong data type is passed, SPDC_WRONG_DTYPE
 * is returned. If the array dimension is > 1, SPDC_NDIM_INCOMPATIBLE is
 * returned.*/
int spdc_linspace(double min, double max, spdc_mdarray * arr)
{
	if (arr->dtype != SPDC_DTYPE_DOUBLE) {
		return (SPDC_WRONG_DTYPE);
	}

	if (arr->ndim != 1) {
		return (SPDC_NDIM_INCOMPATIBLE);
	}

	size_t N = arr->shape[0];
	double *data = NULL;
	data = (double *)arr->data;

	for (size_t i = 0; i < N; i++) {
		data[i] = min + i * (max - min) / ((double)(N - 1));
	}
	return (SPDC_RETURN_OK);
}

int spdc_zeros(spdc_mdarray * arr)
{
	size_t N = spdc_get_size(arr);
	if (arr->dtype == SPDC_DTYPE_DOUBLE) {
		for (ptrdiff_t i = 0; i < N; i++) {
			SPDC_GET_DOUBLE(*arr, i) = 0.0;
		}
	} else if (arr->dtype == SPDC_DTYPE_COMPLEX) {
		for (ptrdiff_t i = 0; i < N; i++) {
			SPDC_GET_COMPLEX(*arr, i) = 0.0 + 0.0 * I;
		}
	} else {
		return (SPDC_WRONG_DTYPE);
	}
	return (SPDC_RETURN_OK);
}

size_t spdc_get_size(const spdc_mdarray * restrict arr)
{
	size_t N = 1;
	for (ptrdiff_t i = 0; i < arr->ndim; i++) {
		N *= arr->shape[i];
	}
	return (N);
}

/* Reduce the array dimension by ax_count by summing over the given axes */
int spdc_sum_reduce(spdc_mdarray * arr, int ax_count, int *axes,
		    spdc_mdarray * res, double normalization, int abs_square)
{
	/* Validate input */
	assert(arr->ndim == res->ndim + ax_count);
	assert(arr->dtype == res->dtype);
	assert(arr->dtype == SPDC_DTYPE_DOUBLE
	       || arr->dtype == SPDC_DTYPE_COMPLEX);
	assert(arr->ndim < SPDC_MAX_REDUCE_NDIM);

	/* Multiindices for accessing the input and output arrays */
	ptrdiff_t multi_idx[SPDC_MAX_REDUCE_NDIM];
	ptrdiff_t multi_idx_res[SPDC_MAX_REDUCE_NDIM];

	/* Determine which axes are summed over and which are not */
	int const_axes[SPDC_MAX_REDUCE_NDIM];
	int c_const = 0, c_var = 0;
	int has_index = 0;
	for (int i = 0; i < arr->ndim; i++) {
		for (int j = 0; j < ax_count; j++) {
			if (i == axes[j])
				has_index = 1;
		}

		if (has_index) {
			c_var++;
			has_index = 0;
		} else {
			const_axes[c_const] = i;
			c_const++;
		}

		assert(c_const <= (arr->ndim - ax_count));
		assert(c_var <= ax_count);
	}
	assert(c_const == (arr->ndim - ax_count));
	assert(c_var == (ax_count));

	/* Set output array to zero */
	spdc_zeros(res);

#pragma omp parallel
	{
		/* Create thread-private output storage */
		spdc_mdarray tmp_array;
		spdc_init_mdarray(res->ndim, res->shape, res->dtype,
				  &tmp_array);
		spdc_zeros(&tmp_array);

		size_t n_items = spdc_get_size(arr);
#pragma omp for private(multi_idx, multi_idx_res) nowait
		for (ptrdiff_t i = 0; i < n_items; i++) {
			/* Visit each element in input array and sum cumulatively */

			/* Get multiindex of input array */
			spdc_unravel(arr, i, multi_idx);

			/* Calculate (flat) target index in output array */
			for (int j = 0; j < res->ndim; j++) {
				multi_idx_res[j] = multi_idx[const_axes[j]];
			}
			ptrdiff_t k = 0;
			spdc_ravel(res, multi_idx_res, &k);

			/* Add to output array */
			if (arr->dtype == SPDC_DTYPE_DOUBLE) {
				double tmp_var = SPDC_GET_DOUBLE(*arr, i);
				if (abs_square) {
					tmp_var = pow(fabs(tmp_var), 2);
				}
				SPDC_GET_DOUBLE(tmp_array, k) +=
				    tmp_var * normalization;

			} else {
				double complex tmp_var =
				    SPDC_GET_COMPLEX(*arr, i);
				if (abs_square) {
					tmp_var = pow(cabs(tmp_var), 2);
				}
				SPDC_GET_COMPLEX(tmp_array, k) +=
				    tmp_var * normalization;
			}

		}

		/* Join subarrays */
#pragma omp critical
		{
			n_items = spdc_get_size(res);
			if (arr->dtype == SPDC_DTYPE_DOUBLE) {
				for (ptrdiff_t i = 0; i < n_items; i++) {
					SPDC_GET_DOUBLE(*res, i) +=
					    SPDC_GET_DOUBLE(tmp_array, i);
				}
			} else {
				for (ptrdiff_t i = 0; i < n_items; i++) {
					SPDC_GET_COMPLEX(*res, i) +=
					    SPDC_GET_COMPLEX(tmp_array, i);
				}
			}
		}

		/* Free temporary storage */
		spdc_free(&tmp_array);
	}

	return (SPDC_RETURN_OK);

}

/* Reduce the array dimension by ax_count by summing over the given axes, The
 * code here employs a double loop and visit each output element separately
 * and thus prevents any concurrent write access into the output array.
 * Therefore the codes scales much better with using multiple threads with
 * openMP.*/
int spdc_sum_reduce_old(spdc_mdarray * arr, int ax_count, int *axes,
			spdc_mdarray * res, double normalization,
			int abs_square)
{
	/* Validate input */
	assert(arr->ndim == res->ndim + ax_count);
	assert(arr->dtype == res->dtype);
	assert(arr->dtype == SPDC_DTYPE_DOUBLE
	       || arr->dtype == SPDC_DTYPE_COMPLEX);

	/* Multiindices for accessing the input and output arrays */
	ptrdiff_t multi_idx_res[SPDC_MAX_REDUCE_NDIM];
	ptrdiff_t multi_idx_arr[SPDC_MAX_REDUCE_NDIM];
	ptrdiff_t multi_idx_sub[SPDC_MAX_REDUCE_NDIM];

	/* Determine which axes are summed over and which are not */
	int *const_axes = malloc(sizeof(int) * (arr->ndim - ax_count));
	int *var_axes = malloc(sizeof(int) * ax_count);
	int c_const = 0, c_var = 0;
	int has_index = 0;
	for (int i = 0; i < arr->ndim; i++) {
		for (int j = 0; j < ax_count; j++) {
			if (i == axes[j])
				has_index = 1;
		}

		if (has_index) {
			var_axes[c_var] = i;
			c_var++;
			has_index = 0;
		} else {
			const_axes[c_const] = i;
			c_const++;
		}

		assert(c_const <= (arr->ndim - ax_count));
		assert(c_var <= ax_count);
	}
	assert(c_const == (arr->ndim - ax_count));
	assert(c_var == (ax_count));

	/* Create fake mdarray without data but correct dimensions for the
	 * subarrays of arr that are summed up to yield each element of the result*/
	spdc_mdarray fake_array;
	fake_array.dtype = arr->dtype;
	fake_array.ndim = ax_count;
	fake_array.shape = (size_t *) malloc(sizeof(size_t) * fake_array.ndim);
	for (int i = 0; i < ax_count; i++) {
		fake_array.shape[i] = arr->shape[var_axes[i]];
	}

	/* Init output array with zeros */
	spdc_zeros(res);

	/* Sum over axes */
	size_t n_items_res = spdc_get_size(res);
	size_t n_items_sub = spdc_get_size(&fake_array);
#pragma omp parallel for private(multi_idx_res, multi_idx_arr, multi_idx_sub)
	for (ptrdiff_t i = 0; i < n_items_res; i++) {
		/* Visit each element in output array */

		/* Compute multi_idx for output array */
		spdc_unravel(res, i, multi_idx_res);

		for (ptrdiff_t j = 0; j < n_items_sub; j++) {
			/* Sum up each element from subarray */

			/* Set constant idx in input array multi_idx_arr */
			for (int k = 0; k < (arr->ndim - ax_count); k++)
				multi_idx_arr[const_axes[k]] = multi_idx_res[k];

			/* Set variable idx in input array multi_idx_arr */
			spdc_unravel(&fake_array, j, multi_idx_sub);
			for (int k = 0; k < ax_count; k++) {
				multi_idx_arr[var_axes[k]] = multi_idx_sub[k];
			}

			/* Convert the complete multi_idx_arr back to a flat idx */
			ptrdiff_t flat_idx = 0;
			spdc_ravel(arr, multi_idx_arr, &flat_idx);

			/* Add cumulatively from input to output array */
			if (arr->dtype == SPDC_DTYPE_DOUBLE) {
				double tmp_var =
				    SPDC_GET_DOUBLE(*arr, flat_idx);
				if (abs_square) {
					tmp_var = pow(fabs(tmp_var), 2);
				}
				SPDC_GET_DOUBLE(*res, i) +=
				    tmp_var * normalization;
			} else {
				double complex tmp_var =
				    SPDC_GET_COMPLEX(*arr, flat_idx);
				if (abs_square) {
					tmp_var = pow(cabs(tmp_var), 2);
				}
				SPDC_GET_COMPLEX(*res, i) +=
				    tmp_var * normalization;
			}

		}
	}

	free(const_axes);
	free(var_axes);
	free(fake_array.shape);

	return (SPDC_RETURN_OK);

}

/* Compute the absolute squared value of the array *in and writes the result
 * to *out */
int spdc_abs_square(spdc_mdarray * in, spdc_mdarray * out)
{
	assert(in->ndim == out->ndim);
	assert(spdc_get_size(in) == spdc_get_size(out));

	size_t n_items = spdc_get_size(in);
	if ((in->dtype == SPDC_DTYPE_DOUBLE)
	    && (out->dtype == SPDC_DTYPE_DOUBLE)) {
#pragma omp parallel for
		for (ptrdiff_t i = 0; i < n_items; i++) {
			SPDC_GET_DOUBLE(*out, i) =
			    pow(fabs(SPDC_GET_DOUBLE(*in, i)), 2);
		}
	} else if ((in->dtype == SPDC_DTYPE_COMPLEX)
		   && (out->dtype == SPDC_DTYPE_DOUBLE)) {
#pragma omp parallel for
		for (ptrdiff_t i = 0; i < n_items; i++) {
			SPDC_GET_DOUBLE(*out, i) =
			    pow(cabs(SPDC_GET_COMPLEX(*in, i)), 2);
		}
	} else if ((in->dtype == SPDC_DTYPE_COMPLEX)
		   && (out->dtype == SPDC_DTYPE_COMPLEX)) {
#pragma omp parallel for
		for (ptrdiff_t i = 0; i < n_items; i++) {
			SPDC_GET_COMPLEX(*out, i) =
			    pow(cabs(SPDC_GET_COMPLEX(*in, i)), 2);
		}
	} else {
		return (SPDC_WRONG_DTYPE);
	}

	return (SPDC_RETURN_OK);

}

/* Initialize an mdarray of dimension (or rank) ndim and shape *shape. Dtype
 * may either be SPDC_DTYPE_DOUBLE or SPDC_DTYPE_COMPLEX for a real or complex
 * array respectively. Memory for the array data is allocated but _not_
 * initialized and consequently contains random values after initialization.*/
int spdc_init_mdarray(int ndim, size_t * shape, int dtype, spdc_mdarray * arr)
{
	size_t N = 0;
	void *data = NULL;

	/* Initialize array struct */
	arr->ndim = ndim;
	arr->dtype = dtype;
	arr->shape = malloc(sizeof(size_t) * ndim);
	memcpy(arr->shape, shape, ndim * sizeof(size_t));

	/* Calculate number of elements in array */
	N = spdc_get_size(arr);

	/* Allocate memory */
	if (arr->dtype == SPDC_DTYPE_COMPLEX) {
		data = fftw_malloc(sizeof(double complex) * N);
	}

	if (arr->dtype == SPDC_DTYPE_DOUBLE) {
		data = fftw_malloc(sizeof(double) * N);
	}

	if (!data) {
		printf("Failed to allocate memory for %i complex elements",
		       (int)N);
		return (-1);
	}

	arr->data = data;
	return (0);

}

int spdc_free(spdc_mdarray * arr)
{
	fftw_free(arr->data);
	free(arr->shape);
	return (SPDC_RETURN_OK);
}
