
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <complex.h>
#include <fftw3.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <assert.h>
#include <omp.h>
#include <ssvd.h>
#include <gsl/gsl_blas.h>

#include <argp.h>

/* Data type definitions*/
#define SPDC_DTYPE_DOUBLE 10
#define SPDC_DTYPE_COMPLEX 20

#define SPDC_MAX_REDUCE_NDIM 20

/* Macro definitions to access mdarrays */
#define SPDC_GET_COMPLEX(A, i) (((complex double*) (A).data)[(i)])
#define SPDC_GET_COMPLEX2(A, i0, i1) ( ((double complex *) (A).data)[(i1) + (A).shape[1] * (i0)] )

#define SPDC_GET_DOUBLE(A, i) (((double*) (A).data)[(i)])
#define SPDC_GET_DOUBLE2(A, i0, i1) ( ((double *) (A).data)[(i1) + (A).shape[1] * (i0)] )

/* Error codes */
#define SPDC_RETURN_OK 0
#define SPDC_GENERIC_ERROR -1
#define SPDC_WRONG_DTYPE -2
#define SPDC_NDIM_INCOMPATIBLE -3
#define SPDC_NOT_POWER_OF_TWO -4
#define SPDC_MEMORY_ALLOC_ERROR -5
#define SPDC_SHAPE_INCOMPATIBLE -6
#define SPDC_PARAMETER_OUT_OF_RANGE -7
#define SPDC_INDEX_OUT_OF_RANGE -7

/* Physical constants */
#define SPDC_C0 2.99792458e8

/* Structure definitions */
typedef struct field_co {
	/* This structure holds all information about an electric field that
	 * participates in an SPDC process.
	 */
	char name[32];
	double n_o, n_e, lambda, k, omega;
	double alpha, beta, gamma, eta;
} field_co;

/* Simple multidimensional array type */
typedef struct spdc_mdarray {
	void *data;
	size_t *shape;
	int ndim;
	int dtype;		/* Ugly hack with 10 == double, 20 == complex */
} spdc_mdarray;

/* SPDC parameter struct */
/*[[[cog
import cog_codegen

options = cog_codegen.get_options()
cog_codegen.generate_structure(options)
]]]*/
struct spdc_prm {
	double lambda_pump;
	double lambda_signal;
	double lambda_idler;
	int hermite_gauss_m;
	int hermite_gauss_n;
	double pump_w0;
	double pump_z;
	int pseudo_tem;
	double width;
	double length;
	double theta;
	int phase;
	int phase_compensation;
	int phase_match_type;
	int N;
	int schmidt_rank;
	int eigenvalues_only;
	double focal_length;
	double nf_signal_offset;
	double nf_idler_offset;
	double ff_signal_offset;
	double ff_idler_offset;
	double knife_edge_pos_pump_px;
	double knife_edge_pos_pump_py;
	int knife_edge_invert_pump_px;
	int knife_edge_invert_pump_py;
	double circ_ap_r_pump_p;
	double circ_ap_cntr_x_pump_p;
	double circ_ap_cntr_y_pump_p;
	double single_slit_cntr_pump_px;
	double single_slit_cntr_pump_py;
	double single_slit_width_pump_px;
	double single_slit_width_pump_py;
	double double_slit_sep_pump_px;
	double double_slit_sep_pump_py;
	double double_slit_width_pump_px;
	double double_slit_width_pump_py;
	double knife_edge_pos_f1_sx;
	double knife_edge_pos_f1_sy;
	int knife_edge_invert_f1_sx;
	int knife_edge_invert_f1_sy;
	double circ_ap_r_f1_s;
	double circ_ap_cntr_x_f1_s;
	double circ_ap_cntr_y_f1_s;
	double single_slit_cntr_f1_sx;
	double single_slit_cntr_f1_sy;
	double single_slit_width_f1_sx;
	double single_slit_width_f1_sy;
	double double_slit_sep_f1_sx;
	double double_slit_sep_f1_sy;
	double double_slit_width_f1_sx;
	double double_slit_width_f1_sy;
	double knife_edge_pos_f1_ix;
	double knife_edge_pos_f1_iy;
	int knife_edge_invert_f1_ix;
	int knife_edge_invert_f1_iy;
	double circ_ap_r_f1_i;
	double circ_ap_cntr_x_f1_i;
	double circ_ap_cntr_y_f1_i;
	double single_slit_cntr_f1_ix;
	double single_slit_cntr_f1_iy;
	double single_slit_width_f1_ix;
	double single_slit_width_f1_iy;
	double double_slit_sep_f1_ix;
	double double_slit_sep_f1_iy;
	double double_slit_width_f1_ix;
	double double_slit_width_f1_iy;
	double knife_edge_pos_f2_sx;
	double knife_edge_pos_f2_sy;
	int knife_edge_invert_f2_sx;
	int knife_edge_invert_f2_sy;
	double circ_ap_r_f2_s;
	double circ_ap_cntr_x_f2_s;
	double circ_ap_cntr_y_f2_s;
	double single_slit_cntr_f2_sx;
	double single_slit_cntr_f2_sy;
	double single_slit_width_f2_sx;
	double single_slit_width_f2_sy;
	double double_slit_sep_f2_sx;
	double double_slit_sep_f2_sy;
	double double_slit_width_f2_sx;
	double double_slit_width_f2_sy;
	double knife_edge_pos_f2_ix;
	double knife_edge_pos_f2_iy;
	int knife_edge_invert_f2_ix;
	int knife_edge_invert_f2_iy;
	double circ_ap_r_f2_i;
	double circ_ap_cntr_x_f2_i;
	double circ_ap_cntr_y_f2_i;
	double single_slit_cntr_f2_ix;
	double single_slit_cntr_f2_iy;
	double single_slit_width_f2_ix;
	double single_slit_width_f2_iy;
	double double_slit_sep_f2_ix;
	double double_slit_sep_f2_iy;
	double double_slit_width_f2_ix;
	double double_slit_width_f2_iy;
	int pump_pump_mask_type;
	int f1_signal_mask_type;
	int f1_idler_mask_type;
	int f2_signal_mask_type;
	int f2_idler_mask_type;
};
/*[[[end]]]*/

/* Macro definitions to ravel / unravel indices. Ravel converts a multiindex
 * into a flat index suitable for direct access to the underlying buffer.
 * Unravel is the inverse operation to generate a multiindex from a position
 * in a linear array. */
#define SPDC_UNRAVEL2(arr, flat_idx, i, j)  {\
                                    size_t N = spdc_get_size(&arr);\
                                    size_t tmp_flat_idx = (flat_idx);\
                                    (i) = tmp_flat_idx / (N / (arr).shape[0]);\
                                    N /= (arr).shape[0];\
                                    tmp_flat_idx -= (i) * N;\
                                    (j) = tmp_flat_idx / (N / (arr).shape[1]);\
                                    }

#define SPDC_UNRAVEL4(arr, flat_idx, i, j, k, l)  {\
                                    size_t N = spdc_get_size(&arr);\
                                    size_t tmp_flat_idx = (flat_idx);\
                                    (i) = tmp_flat_idx / (N / (arr).shape[0]);\
                                    N /= (arr).shape[0];\
                                    tmp_flat_idx -= (i) * N;\
                                    (j) = tmp_flat_idx / (N / (arr).shape[1]);\
                                    N /= (arr).shape[1];\
                                    tmp_flat_idx -= (j) * N;\
                                    (k) = tmp_flat_idx / (N / (arr).shape[2]);\
                                    N /= (arr).shape[2];\
                                    tmp_flat_idx -= (k) * N;\
                                    (l) = tmp_flat_idx / (N / (arr).shape[3]);\
                                    }

/* Function declarations */

int spdc_walkoff_coefficients(double n_o, double n_e, double theta,
			      field_co * co);
int spdc_sellmeier_bbo(double l, double *n_o, double *n_e);
int spdc_gauss_hermite(int m, int n, spdc_mdarray * x, spdc_mdarray * y,
		       double z, double w_0, double lambda, spdc_mdarray * res);
int spdc_pseudo_tem10(spdc_mdarray * x, spdc_mdarray * y, double w_0,
		      double lambda, spdc_mdarray * res);
int spdc_index_ellipsoid(double n_o, double n_e, double theta, double *n);
double complex spdc_hermite_H(int n, double complex x);
inline double spdc_gaussian(double x, double sigma, double mu);
int spdc_parse_args(struct argp *argp, int argc, char **argv,
		    struct spdc_prm *prm);
int spdc_dump_prm(struct spdc_prm prm);
int spdc_dump_array(spdc_mdarray * arr);
int spdc_sum_reduce(spdc_mdarray * arr, int ax_count, int *axes,
		    spdc_mdarray * res, double normalization, int abs_square);
int spdc_abs_square(spdc_mdarray * in, spdc_mdarray * out);
int spdc_fftshift(spdc_mdarray * arr);
int spdc_fftfreq(spdc_mdarray * arr, double sampling_period);
inline int spdc_ravel(const spdc_mdarray * arr, const ptrdiff_t * multi_idx,
		      ptrdiff_t * flat_idx);
inline int spdc_unravel(const spdc_mdarray * arr, ptrdiff_t flat_idx,
			ptrdiff_t * multi_idx);
inline double spdc_get_double(const spdc_mdarray * arr, ptrdiff_t * n);
inline double complex spdc_get_complex(const spdc_mdarray * arr, ptrdiff_t * n);
size_t spdc_get_size(const spdc_mdarray * restrict arr);
int spdc_zeros(spdc_mdarray * arr);
int spdc_linspace(double min, double max, spdc_mdarray * arr);
int spdc_init_mdarray(int ndim, size_t * shape, int dtype, spdc_mdarray * arr);
int spdc_free(spdc_mdarray * arr);

/*--------------------------------------------------------------------------*/
/* Function definitions for inline functions */
/*--------------------------------------------------------------------------*/

/*[[[cog
import cog_codegen

cog_codegen.generate_mask_functions()
]]]*/
inline double complex spdc_knife_edge(double c1, double pos, const int invert)
{
	double mask_factor = 1.0;
	if (invert == 1) {
		pos = -pos;
		c1 = -c1;
	}
	if (c1 < pos) {
		mask_factor = 0.0;
	}
	return (mask_factor);
}

inline double complex spdc_circular_aperture(const double c1, const double c2,
					     const double r,
					     const double cntr_x,
					     const double cntr_y)
{
	double mask_factor = 1.0;
	const double d1 = c1 - cntr_x;
	const double d2 = c2 - cntr_y;
	if ((d1 * d1 + d2 * d2) > (r * r)) {
		mask_factor = 0;
	}
	return (mask_factor);
}

inline double complex spdc_single_slit(const double c1, const double cntr,
				       const double width)
{
	double mask_factor = 1.0;
	if (fabs(c1 - cntr) > width / 2.0) {
		mask_factor = 0.0;
	}
	return (mask_factor);
}

inline double complex spdc_double_slit(const double c1, const double sep,
				       const double width)
{
	return (spdc_single_slit(fabs(c1), sep / 2.0, width));
}

///[[[end]]]

/* Generic masking function for the pump beam */
inline int spdc_mask_2d(const spdc_mdarray * restrict c1,
			const spdc_mdarray * restrict c2, struct
			spdc_mdarray *restrict a, const struct spdc_prm *prm)
{
	assert(a->dtype == SPDC_DTYPE_COMPLEX);
	assert(c1->dtype == SPDC_DTYPE_DOUBLE);
	assert(c2->dtype == SPDC_DTYPE_DOUBLE);
	assert(a->ndim == 2);
	const size_t n_items = spdc_get_size(a);
	double complex *restrict data = (double complex *)a->data;
	const double *restrict c1_data = (double *)c1->data;
	const double *restrict c2_data = (double *)c2->data;

#pragma omp parallel for
	for (ptrdiff_t i = 0; i < n_items; i++) {
		ptrdiff_t j, k;
		SPDC_UNRAVEL2(*a, i, j, k);
		/*[[[cog
		   import cog_codegen
		   cog_codegen.generate_pump_mask_selection_code()
		   ]]] */
		if (prm->pump_pump_mask_type == 1) {
			data[i] *=
			    spdc_knife_edge(c1_data[j],
					    prm->knife_edge_pos_pump_px,
					    prm->knife_edge_invert_pump_px);
			data[i] *=
			    spdc_knife_edge(c2_data[k],
					    prm->knife_edge_pos_pump_py,
					    prm->knife_edge_invert_pump_py);
		}
		if (prm->pump_pump_mask_type == 2) {
			data[i] *=
			    spdc_circular_aperture(c1_data[j], c2_data[k],
						   prm->circ_ap_r_pump_p,
						   prm->circ_ap_cntr_x_pump_p,
						   prm->circ_ap_cntr_y_pump_p);
		}
		if (prm->pump_pump_mask_type == 3) {
			data[i] *=
			    spdc_single_slit(c1_data[j],
					     prm->single_slit_cntr_pump_px,
					     prm->single_slit_width_pump_px);
			data[i] *=
			    spdc_single_slit(c2_data[k],
					     prm->single_slit_cntr_pump_py,
					     prm->single_slit_width_pump_py);
		}
		if (prm->pump_pump_mask_type == 4) {
			data[i] *=
			    spdc_double_slit(c1_data[j],
					     prm->double_slit_sep_pump_px,
					     prm->double_slit_width_pump_px);
			data[i] *=
			    spdc_double_slit(c2_data[k],
					     prm->double_slit_sep_pump_py,
					     prm->double_slit_width_pump_py);
		}
		//[[[end]]]

	}
	return (SPDC_RETURN_OK);
}

inline int spdc_mask(const spdc_mdarray * restrict c1,
		     const spdc_mdarray * restrict c2,
		     const spdc_mdarray * restrict c3,
		     const spdc_mdarray * restrict c4, struct spdc_mdarray
		     *restrict a, const int plane, const struct spdc_prm *prm)
{
	assert(a->dtype == SPDC_DTYPE_COMPLEX);
	assert(c1->dtype == SPDC_DTYPE_DOUBLE);
	assert(c2->dtype == SPDC_DTYPE_DOUBLE);
	assert(a->ndim == 4);
	const size_t n_items = spdc_get_size(a);
	double complex *restrict data = (double complex *)a->data;
	const double *restrict c1_data = (double *)c1->data;
	const double *restrict c2_data = (double *)c2->data;
	const double *restrict c3_data = (double *)c3->data;
	const double *restrict c4_data = (double *)c4->data;

#pragma omp parallel for
	for (ptrdiff_t i = 0; i < n_items; i++) {
		ptrdiff_t j, k, l, m;
		SPDC_UNRAVEL4(*a, i, j, k, l, m);
		/*[[[cog
		   import cog_codegen
		   cog_codegen.generate_mask_selection_code()
		   ]]] */
		if (plane == 1) {
			if (prm->f1_signal_mask_type == 1) {
				data[i] *=
				    spdc_knife_edge(c1_data[j],
						    prm->knife_edge_pos_f1_sx,
						    prm->
						    knife_edge_invert_f1_sx);
				data[i] *=
				    spdc_knife_edge(c2_data[k],
						    prm->knife_edge_pos_f1_sy,
						    prm->
						    knife_edge_invert_f1_sy);
			}
			if (prm->f1_signal_mask_type == 2) {
				data[i] *=
				    spdc_circular_aperture(c1_data[j],
							   c2_data[k],
							   prm->circ_ap_r_f1_s,
							   prm->
							   circ_ap_cntr_x_f1_s,
							   prm->
							   circ_ap_cntr_y_f1_s);
			}
			if (prm->f1_signal_mask_type == 3) {
				data[i] *=
				    spdc_single_slit(c1_data[j],
						     prm->
						     single_slit_cntr_f1_sx,
						     prm->
						     single_slit_width_f1_sx);
				data[i] *=
				    spdc_single_slit(c2_data[k],
						     prm->
						     single_slit_cntr_f1_sy,
						     prm->
						     single_slit_width_f1_sy);
			}
			if (prm->f1_signal_mask_type == 4) {
				data[i] *=
				    spdc_double_slit(c1_data[j],
						     prm->double_slit_sep_f1_sx,
						     prm->
						     double_slit_width_f1_sx);
				data[i] *=
				    spdc_double_slit(c2_data[k],
						     prm->double_slit_sep_f1_sy,
						     prm->
						     double_slit_width_f1_sy);
			}
			if (prm->f1_idler_mask_type == 1) {
				data[i] *=
				    spdc_knife_edge(c3_data[l],
						    prm->knife_edge_pos_f1_ix,
						    prm->
						    knife_edge_invert_f1_ix);
				data[i] *=
				    spdc_knife_edge(c4_data[m],
						    prm->knife_edge_pos_f1_iy,
						    prm->
						    knife_edge_invert_f1_iy);
			}
			if (prm->f1_idler_mask_type == 2) {
				data[i] *=
				    spdc_circular_aperture(c3_data[l],
							   c4_data[m],
							   prm->circ_ap_r_f1_i,
							   prm->
							   circ_ap_cntr_x_f1_i,
							   prm->
							   circ_ap_cntr_y_f1_i);
			}
			if (prm->f1_idler_mask_type == 3) {
				data[i] *=
				    spdc_single_slit(c3_data[l],
						     prm->
						     single_slit_cntr_f1_ix,
						     prm->
						     single_slit_width_f1_ix);
				data[i] *=
				    spdc_single_slit(c4_data[m],
						     prm->
						     single_slit_cntr_f1_iy,
						     prm->
						     single_slit_width_f1_iy);
			}
			if (prm->f1_idler_mask_type == 4) {
				data[i] *=
				    spdc_double_slit(c3_data[l],
						     prm->double_slit_sep_f1_ix,
						     prm->
						     double_slit_width_f1_ix);
				data[i] *=
				    spdc_double_slit(c4_data[m],
						     prm->double_slit_sep_f1_iy,
						     prm->
						     double_slit_width_f1_iy);
			}
		}
		if (plane == 2) {
			if (prm->f2_signal_mask_type == 1) {
				data[i] *=
				    spdc_knife_edge(c1_data[j],
						    prm->knife_edge_pos_f2_sx,
						    prm->
						    knife_edge_invert_f2_sx);
				data[i] *=
				    spdc_knife_edge(c2_data[k],
						    prm->knife_edge_pos_f2_sy,
						    prm->
						    knife_edge_invert_f2_sy);
			}
			if (prm->f2_signal_mask_type == 2) {
				data[i] *=
				    spdc_circular_aperture(c1_data[j],
							   c2_data[k],
							   prm->circ_ap_r_f2_s,
							   prm->
							   circ_ap_cntr_x_f2_s,
							   prm->
							   circ_ap_cntr_y_f2_s);
			}
			if (prm->f2_signal_mask_type == 3) {
				data[i] *=
				    spdc_single_slit(c1_data[j],
						     prm->
						     single_slit_cntr_f2_sx,
						     prm->
						     single_slit_width_f2_sx);
				data[i] *=
				    spdc_single_slit(c2_data[k],
						     prm->
						     single_slit_cntr_f2_sy,
						     prm->
						     single_slit_width_f2_sy);
			}
			if (prm->f2_signal_mask_type == 4) {
				data[i] *=
				    spdc_double_slit(c1_data[j],
						     prm->double_slit_sep_f2_sx,
						     prm->
						     double_slit_width_f2_sx);
				data[i] *=
				    spdc_double_slit(c2_data[k],
						     prm->double_slit_sep_f2_sy,
						     prm->
						     double_slit_width_f2_sy);
			}
			if (prm->f2_idler_mask_type == 1) {
				data[i] *=
				    spdc_knife_edge(c3_data[l],
						    prm->knife_edge_pos_f2_ix,
						    prm->
						    knife_edge_invert_f2_ix);
				data[i] *=
				    spdc_knife_edge(c4_data[m],
						    prm->knife_edge_pos_f2_iy,
						    prm->
						    knife_edge_invert_f2_iy);
			}
			if (prm->f2_idler_mask_type == 2) {
				data[i] *=
				    spdc_circular_aperture(c3_data[l],
							   c4_data[m],
							   prm->circ_ap_r_f2_i,
							   prm->
							   circ_ap_cntr_x_f2_i,
							   prm->
							   circ_ap_cntr_y_f2_i);
			}
			if (prm->f2_idler_mask_type == 3) {
				data[i] *=
				    spdc_single_slit(c3_data[l],
						     prm->
						     single_slit_cntr_f2_ix,
						     prm->
						     single_slit_width_f2_ix);
				data[i] *=
				    spdc_single_slit(c4_data[m],
						     prm->
						     single_slit_cntr_f2_iy,
						     prm->
						     single_slit_width_f2_iy);
			}
			if (prm->f2_idler_mask_type == 4) {
				data[i] *=
				    spdc_double_slit(c3_data[l],
						     prm->double_slit_sep_f2_ix,
						     prm->
						     double_slit_width_f2_ix);
				data[i] *=
				    spdc_double_slit(c4_data[m],
						     prm->double_slit_sep_f2_iy,
						     prm->
						     double_slit_width_f2_iy);
			}
		}
		//[[[end]]]
	}

}

inline double spdc_gaussian(double x, double sigma, double mu)
{
	return (exp(-pow(x - mu, 2) / (2.0 * pow(sigma, 2))));
}

/* Get element at position given by multiindex n from real mdarray*/
inline double spdc_get_double(const spdc_mdarray * arr, ptrdiff_t * n)
{
	assert(arr->dtype == SPDC_DTYPE_DOUBLE);
	ptrdiff_t idx = 0;
	spdc_ravel(arr, n, &idx);
	return (((double *)
		 arr->data)[idx]);
}

/* Get element at position given by multiindex n from complex mdarray*/
inline double complex spdc_get_complex(const spdc_mdarray * arr, ptrdiff_t * n)
{
	assert(arr->dtype == SPDC_DTYPE_COMPLEX);
	ptrdiff_t idx = 0;
	spdc_ravel(arr, n, &idx);
	return (((double complex *)
		 arr->data)[idx]);
}

/* A multiindex *n is converted to a flat index suitable to directly index the
 * data buffer of an mdarray struct with the same shape as arr. The argument n
 * must be a pointer to a buffer containing arr->ndim ptrdiff_t integers
 * that constitute a valid multiindx (e.g. 0 < n[i] < arr->shape[i]. The
 * result is written into the memory location pointed to by idx which must be
 * big enough to hold exactly one ptrdiff_t element.*/
inline int spdc_ravel(const spdc_mdarray * arr, const ptrdiff_t * multi_idx,
		      ptrdiff_t * flat_idx)
{
	*flat_idx = 0;
	for (int i = 0; i < arr->ndim - 1; i++) {
		if (multi_idx[i] >= arr->shape[i]) {
			return (SPDC_INDEX_OUT_OF_RANGE);
		}
		*flat_idx = ((*flat_idx) + multi_idx[i])
		    * arr->shape[i + 1];
	}
	(*flat_idx) += multi_idx[arr->ndim - 1];
	return (SPDC_RETURN_OK);
}

/* A flat index flat_idx is converted to an n-multiindex suitable to index an
 * n-dimensiona mdarray with the shape of arr. idx must point to a buffer with
 * enough space to hold exactly n ptrdiff_t variables.*/
inline int spdc_unravel(const spdc_mdarray * arr, ptrdiff_t flat_idx,
			ptrdiff_t * multi_idx)
{
	assert(multi_idx != NULL);
	size_t N = 0;
	N = spdc_get_size(arr);

	/* Check that the index is valid */
	if (flat_idx >= N) {
		return (SPDC_INDEX_OUT_OF_RANGE);
	}

	for (int i = 0; i < arr->ndim; i++) {
		multi_idx[i] = flat_idx / (N / arr->shape[i]);
		N = N / arr->shape[i];
		flat_idx -= multi_idx[i] * N;
	}
	return (SPDC_RETURN_OK);
}
