
#include <stdlib.h>
#include <check.h>
#include <time.h>
#include "../src/spdcc.h"

START_TEST(test_sellmeier_bbo)
{
    /* Test correct calculation of the ordinary and extraordinary refractive
     * indices at 400nm
     */
    double n_o, n_e, l;
    int ret;
    l = 400e-9;
    ret = spdc_sellmeier_bbo(l, &n_o, &n_e);
    ck_assert(ret == 0);
    ck_assert(abs(n_o - 1.69337) < 1e-5);
    ck_assert(abs(n_e - 1.56874) < 1e-5);
}
END_TEST

START_TEST(test_index_ellipsoid)
{
    /* Test correct calculation of the index ellipsoid given n_e and n_o */
    double n_o, n_e, l;
    l = 400e-9;
    spdc_sellmeier_bbo(l, &n_o, &n_e);
    ck_assert(0);

}
END_TEST

START_TEST(test_gauss_hermite)
{
    /* Test correct calculation of a Hermite-Gauss laser mode */
    int m=1, n=0, retcode;
    size_t x_shape[1], res_shape[2], N=1024;
    double z=0.0, w_0=1e-3, lambda=400e-9, L=3e-3, x_tmp, y_tmp, intensity, intensity_numeric;
    ptrdiff_t multi_idx[2], rand_idx[2];
    spdc_mdarray x, y, res;


    /* Array initialization */
    x_shape[0] = N;
    res_shape[0] = N; res_shape[1] = N;
    spdc_init_mdarray(1, x_shape, SPDC_DTYPE_DOUBLE, &x);
    spdc_init_mdarray(1, x_shape, SPDC_DTYPE_DOUBLE, &y);
    spdc_init_mdarray(2, res_shape, SPDC_DTYPE_COMPLEX, &res);

    /* Fill x- and y-arrays */
    spdc_linspace(-L, L, &x);
    spdc_linspace(-L, L, &y);

    /* Calculate field distribution for Gauss-Hermite mode */
    retcode = spdc_gauss_hermite(m, n, &x, &y, z, w_0, lambda, &res);
    ck_assert(retcode == SPDC_RETURN_OK);

    /* Check for expected field distribution at randomly selected points */
    srand(time(NULL));
    for (ptrdiff_t i=0; i < 100; i++) {
        rand_idx[0] = rand() % N;
        rand_idx[1] = rand() % N;
        x_tmp = spdc_get_double(&x, rand_idx);
        y_tmp = spdc_get_double(&y, (rand_idx + 1));
        intensity = pow(cabs(spdc_hermite_H(m, sqrt(2) * x_tmp / w_0) *
                exp(-pow(x_tmp / w_0, 2))), 2) * pow(cabs(spdc_hermite_H(n,
                sqrt(2) * y_tmp / w_0) * exp(-pow(y_tmp / w_0, 2))), 2);
        intensity_numeric = cabs(spdc_get_complex(&res, rand_idx));
        ck_assert(fabs(intensity - pow(intensity_numeric, 2)) <= 5 * DBL_EPSILON);
    }
}
END_TEST

START_TEST(test_hermite_polynomials)
{
    /* TODO: Needs more comprehensive testing code */
    double complex x = 2.0 + I, y;
    y = spdc_hermite_H(1, x);
    ck_assert(fabs(creal(y) - 4.0) <= DBL_EPSILON);
    ck_assert(fabs(cimag(y) - 2.0) <= DBL_EPSILON);
}
END_TEST

/*----------------------------------------------------------------------------
 * Tests for array manipulation and initialization
 */

START_TEST(test_array_allocation)
{
    size_t n = 128;
    int ndim = 3;
    size_t shape[3] = {n, n + 2, n + 4};
    spdc_mdarray arr;

    spdc_init_mdarray(ndim, shape, SPDC_DTYPE_COMPLEX, &arr);

    /* Check that the array dimensions and shape are correct */
    ck_assert(arr.ndim == ndim);
    ck_assert(arr.shape[0] == n);
    ck_assert(arr.shape[1] == n + 2);
    ck_assert(arr.shape[2] == n + 4);

    /* Check that the shape array is properly copied */
    shape[1] = 5;
    ck_assert(arr.shape[1] == n + 2);

    /* Check for correct array size */

}
END_TEST

START_TEST(test_ravel)
{
    const int ndim = 3;
    size_t shape[3] = {3, 4, 5};
    spdc_mdarray arr;
    spdc_init_mdarray(ndim, shape, SPDC_DTYPE_DOUBLE, &arr);

    /* Ravel multiindex and check for correctness */
    ptrdiff_t multi_idx[3] = {2, 3, 4};
    ptrdiff_t flat_idx;
    spdc_ravel(&arr, multi_idx, &flat_idx);
    ck_assert(flat_idx == 59);

}
END_TEST

START_TEST(test_unravel)
{
    const int ndim = 3;
    size_t shape[3] = {3, 4, 5};
    spdc_mdarray arr;
    spdc_init_mdarray(ndim, shape, SPDC_DTYPE_DOUBLE, &arr);

    /* Unravel flat index to multiindex */
    ptrdiff_t flat_idx = 59;
    ptrdiff_t multi_idx[3];
    spdc_unravel(&arr, flat_idx, multi_idx);
    ck_assert(multi_idx[0] == 2);
    ck_assert(multi_idx[1] == 3);
    ck_assert(multi_idx[2] == 4);

    /* Check for some index at the beginning of the array */
    flat_idx = 21;
    spdc_unravel(&arr, flat_idx, multi_idx);
    ck_assert(multi_idx[0] == 1);
    ck_assert(multi_idx[1] == 0);
    ck_assert(multi_idx[2] == 1);

}
END_TEST

START_TEST(test_linspace)
{
    ck_assert(1);
}
END_TEST

START_TEST(test_get)
{
    ck_assert(1);
}
END_TEST

START_TEST(test_set)
{
    ck_assert(1);
}
END_TEST

START_TEST(test_sum_reduce)
{
    /* Create arrays */
    size_t N = 2;
    spdc_mdarray arr_2d, arr_3d;
    spdc_mdarray arr_res_1, arr_res_2;
    size_t shape_2d[2] = {N, N + 1}, shape_3d[3] = {N, N + 1, N},
           shape_res_1[1] = {N}, shape_res_2[2] = {N + 1};

    spdc_init_mdarray(2, shape_2d, SPDC_DTYPE_DOUBLE, &arr_2d);
    spdc_init_mdarray(3, shape_3d, SPDC_DTYPE_DOUBLE, &arr_3d);
    spdc_init_mdarray(1, shape_res_1, SPDC_DTYPE_DOUBLE, &arr_res_1);
    spdc_init_mdarray(1, shape_res_2, SPDC_DTYPE_DOUBLE, &arr_res_2);

    for (ptrdiff_t i=0; i < spdc_get_size(&arr_2d); i++) {
        SPDC_GET_DOUBLE(arr_2d, i) = i;
    }

    for (ptrdiff_t i=0; i < spdc_get_size(&arr_3d); i++) {
        SPDC_GET_DOUBLE(arr_3d, i) = i;
    }


    /* Check if sum reduce works correctly for a 2D array while summing over
     * different axes*/
    int sum_axes[1] = {1};
    spdc_sum_reduce(&arr_2d, 1, sum_axes, &arr_res_1, 1.0, 0); 
    sum_axes[0] = 0;
    spdc_sum_reduce(&arr_2d, 1, sum_axes, &arr_res_2, 1.0, 0); 

    ck_assert(SPDC_GET_DOUBLE(arr_res_1, 0) == 3);
    ck_assert(SPDC_GET_DOUBLE(arr_res_1, 1) == 12);

    ck_assert(SPDC_GET_DOUBLE(arr_res_2, 0) == 3);
    ck_assert(SPDC_GET_DOUBLE(arr_res_2, 1) == 5);
    ck_assert(SPDC_GET_DOUBLE(arr_res_2, 2) == 7);

    /* Test if summation with the abs_squared flags works as expected */
    spdc_sum_reduce(&arr_2d, 1, sum_axes, &arr_res_2, 1.0, 1);
    ck_assert(SPDC_GET_DOUBLE(arr_res_2, 0) == 9);


    /* Test if simultaneous summation over two axes works */
    int sum_axes_2[2] = {1, 2};
    spdc_sum_reduce(&arr_3d, 2, sum_axes_2, &arr_res_1, 1.0, 0);

    ck_assert(SPDC_GET_DOUBLE(arr_res_1, 0) == 15);
    ck_assert(SPDC_GET_DOUBLE(arr_res_1, 1) == 51);


    /* Test if dtype constraints are honored */
    
    /* Test if dimensionality constraints are honored */
}
END_TEST

/*----------------------------------------------------------------------------
 * Testing FFTW interoperability
 */

START_TEST(test_fftw_iop)
{
    /* Allocate complex fftw array */
    spdc_mdarray in;
    fftw_plan p;
    size_t N = 1024;
    size_t shape[2] = {N, N}; int shape_int[2] = {N, N};
    int ndim = 2;

    /* Init array */
    spdc_init_mdarray(ndim, shape, SPDC_DTYPE_COMPLEX, &in);
    for (size_t i=0; i < N; i++) {
        ((complex double *) in.data)[i] = 1 + i * I;
    }
    p = fftw_plan_dft(2, shape_int, in.data, in.data, FFTW_FORWARD, FFTW_ESTIMATE);
    fftw_execute(p);
    fftw_destroy_plan(p);
    ck_assert(fabs(creal(SPDC_GET_COMPLEX(in, 0)) - 1024) < DBL_EPSILON);
}
END_TEST


/*----------------------------------------------------------------------------
 * Test suite setup
 */

Suite *
spdc_suite(void)
{
    /* Create test suite */
    Suite *s = suite_create("SPDCc");

    /* Create test cases for calculation of physical helper variables */
    TCase *tc_sellmeier = tcase_create("Sellmeier");
    tcase_add_test(tc_sellmeier, test_sellmeier_bbo);

    /* Test helper functions */
    TCase *tc_helpers = tcase_create("Helper functions");
    tcase_add_test(tc_helpers, test_gauss_hermite);
    tcase_add_test(tc_helpers, test_hermite_polynomials);

    /* Create array test case */
    TCase *tc_mdarray = tcase_create("MD Array");
    tcase_add_test(tc_mdarray, test_array_allocation);
    tcase_add_test(tc_mdarray, test_ravel);
    tcase_add_test(tc_mdarray, test_unravel);
    tcase_add_test(tc_mdarray, test_linspace);
    tcase_add_test(tc_mdarray, test_get);
    tcase_add_test(tc_mdarray, test_set);
    tcase_add_test(tc_mdarray, test_sum_reduce);

    /* Create FFTW test case */
    TCase *tc_fftw = tcase_create("FFTW");
    tcase_add_test(tc_fftw, test_fftw_iop);

    /* Add test cases to suite */
    suite_add_tcase(s, tc_sellmeier);
    suite_add_tcase(s, tc_mdarray);
    suite_add_tcase(s, tc_fftw);
    suite_add_tcase(s, tc_helpers);

    return(s);
}

int
main(void)
{
    int number_failed;

    /* Preparing test suite and runner*/
    Suite *s = spdc_suite();
    SRunner *sr = srunner_create(s);
    srunner_run_all(sr, CK_NORMAL);

    number_failed = srunner_ntests_failed(sr);
    srunner_free(sr);
    return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
